# cointRegSys is a function that calculates two different estimators for
# a homogeneous and cross-sectionally independent system of cointegrating 
# regression. Computed are a pooled fully modified OLS (PFM-OLS) estimator for
# homogeneous parameter vectors and an equation-wise or individual fully 
# modified OLS (IFM-OLS) estimator which is in fact just a fully modified 
# estimator applied on each cointegrating regression seperately.

# The function uses some internal functions from the R package 'cointReg' and
# code snippets from the cointReg functions of the package. 
# Philipp Aschersleben and Martin Wagner (2016). cointReg: Parameter
# Estimation and Inference in a Cointegrating Regression. R package version
# 0.2.0. https://CRAN.R-project.org/package=cointReg

# Depends on package 'cointReg'.

# Arguments
# x         : a numeric matrix with k * N columns of k I(1) regressors
# y         : a numeric matrix with N columns of N I(1) dependant variables
# deter     : a numeric matrix with p columns of p deterministic regressors
# k         : integer, number of I(1) regressors per cointegrating regression
# N         : integer, number of cointegrating regressions
# kernel    : character, any of "ba", "pa", "qs" or "tr", used for kernel based
#             long-run (co-)variance estimation
# bandwidth : character any of "and" or "nw", the bandwidth rule used for
#             kernel estimation
# demeaning : logical, indicates whether or not the series of residuals should
#             be demeaned before computing the long-run (co-)variances
# ...       : additional arguments passed to the function getLongRunVar


# Value 
# The function returns a named list with the following entries
# theta.ifm          : a (p + k) by N numeric matrix containing the individual
#                      FM-OLS estimates for each equation separately
# theta.pfm          : a (p + k) by 1 numeric matrix containing the PFM-OLS 
#                      estimate
# Omegavv.inv.vu.ifm : a k by N matrix containing the matrix product
#                      inverse of \hat{\Omega}^i,i_vv times \hat{\Omega}^i,i_vu
#                      used for fully modifying the cointegrating system in
#                      its i-th column
# Omegavv.inv.vu     : a k by 1 matrix containing the matrix product
#                      inverse of \hat{\Omega}^1,1_vv times \hat{\Omega}^1,1_vu
#                      used for fully modifying the cointegrating system
# omega.u.v2         : the pooled conditional long-run variances
# omega.u.v2.ifm     : the vector of individually estimated conditional long-run
#                      variances

# Author
# Etienne Theising

cointRegSys = function(x, y, deter, k, N, kernel = c("ba", "pa", "qs", "tr"), 
                       bandwidth = c("and", "nw"), demeaning = FALSE, ...) 
{
  require("cointReg")
  p = ncol(deter)
  Te = nrow(x)
  deter = matrix(deter, ncol = p, nrow = Te)
  Omega.star = matrix(0, nrow = k + 1, ncol = k + 1)
  Delta.star = matrix(0, nrow = k + 1, ncol = k + 1)
  y.plus.ifm = matrix(0, nrow = Te, ncol = N)
  u.plus.ifm = matrix(0, nrow = Te, ncol = N)
  theta.ifm = matrix(0, nrow = p + k, ncol = N)
  Omegavv.inv.vu.ifm = matrix(0, nrow = k, ncol = N)
  omega.u.v2.ifm = numeric(N)
  for (n in 1:N) {
    Z = cbind(deter, x[, ((n - 1) * k + 1):(n * k)])
    mod.ols = lm(y[, n] ~ Z - 1)
    u.ols = mod.ols$residuals
    u.4var = u.ols[-1]
    x.delta = matrixStats::colDiffs(x[, ((n - 1) * k + 1):(n * k)])
    u = cbind(u.4var, x.delta)
    if (!is.numeric(bandwidth)) {
      bw = cointReg::getBandwidth(u, bandwidth = bandwidth, kernel = kernel, 
                                  check = FALSE)
      band = switch(bandwidth, and = "Andrews", nw = "Newey-West")
    }
    else {
      bw = bandwidth
      band = "set by user"
    }
    lrvar <- cointReg::getLongRunVar(u, kernel = kernel, bandwidth = bw, 
                                     demeaning = demeaning, check = FALSE, ...)
    Omega.star = Omega.star + lrvar[[1]] / N
    Delta.star = Delta.star + lrvar[[2]] / N
    tmp <- lapply(lrvar, function(x) {
      out <- list()
      out[["all"]] = x
      out[["uu"]] = x[1, 1, drop = FALSE]
      out[["uv"]] = x[1, 2:(1 + k), drop = FALSE]
      out[["vu"]] = x[2:(1 + k), 1, drop = FALSE]
      out[["vv"]] = x[2:(1 + k), 2:(1 + k), drop = FALSE]
      return(out)
    })
    Omega = tmp[[1]]
    Delta = tmp[[2]]
    Omegavv.inv = cointReg:::trySolve(Omega$vv)
    Omegavv.inv.vu = Omegavv.inv %*% Omega$vu
    Omegavv.inv.vu.ifm[, n] = as.numeric(Omegavv.inv.vu)
    omega.u.v2.ifm[n] = Omega$uu - drop(Omega$uv %*% Omegavv.inv.vu)
    Delta.vuplus = Delta$vu - (Delta$vv %*% Omegavv.inv.vu)
    y.plus.ifm[-1, n] = y[-1, n] - (x.delta %*% Omegavv.inv.vu)
    Zfm = Z[-1, , drop = FALSE]
    Zfm2s = cointReg:::trySolve(t(Zfm) %*% Zfm)
    numerat = t(y.plus.ifm[-1, n]) %*% 
      Zfm - cbind(matrix(0, nrow = 1, ncol = p), Te * t(Delta.vuplus))
    theta.ifm[, n] = drop(Zfm2s %*% t(numerat))
    u.plus.ifm[, n] = c(0, y.plus.ifm[-1, n] - Zfm %*% theta.ifm[, n])
  }
  Omegavv.inv = cointReg:::trySolve(Omega.star[2:3, 2:3])
  Omegavv.inv.vu = Omegavv.inv %*% Omega.star[2:3, 1]
  delta.plus.vu = Delta.star[2:3, 1] - (Delta.star[2:3, 2:3] %*% Omegavv.inv.vu)
  omega.u.v2 = Omega.star[1, 1] - drop(Omega.star[1, 2:3] %*% Omegavv.inv.vu)
  y.plus.pfm = matrix(0, nrow = Te, ncol = N)
  for (n in 1:N) {
    x.delta = matrixStats::colDiffs(x[, ((n - 1) * k + 1):(n * k)])
    y.plus.pfm[-1, n] = y[-1, n] - (x.delta %*% Omegavv.inv.vu)
  }
  Z = t(matrix(1, nrow = N) %x% deter[2:Te, ])
  for (i in 1:k) {
    Z = rbind(Z, as.numeric(x[-1, seq(from = i, by = k, length.out = N)]))
  }
  numerat = Z %*% matrix(y.plus.pfm[-1, ], ncol = 1) - 
    N * Te * rbind(matrix(0, nrow = p, ncol = 1), delta.plus.vu)
  theta.pfm = cointReg:::trySolve(Z %*% t(Z)) %*% numerat
  out = list(theta.ifm = theta.ifm, theta.pfm = theta.pfm,
             Omegavv.inv.vu.ifm = Omegavv.inv.vu.ifm,
             Omegavv.inv.vu = Omegavv.inv.vu, omega.u.v2 = omega.u.v2,
             omega.u.v2.ifm = omega.u.v2.ifm)
  return(out)
}
