# cointRegSys is a function that calculates two different estimators for a
# homogeneous or heterogeneous system of cointegrating regression with arbitrary
# error dependence. Computed are a pooled fully modified GLS (PFM-GLS) estimator
# for homogeneous parameter vectors and a fully modified SUR (FM-SUR) estimator 
# for heterogeneous parameter vectors.

# The function uses some internal functions from the R package 'cointReg' and
# code snippets from the cointReg functions of the package. 
# Philipp Aschersleben and Martin Wagner (2016). cointReg: Parameter
# Estimation and Inference in a Cointegrating Regression. R package version
# 0.2.0. https://CRAN.R-project.org/package=cointReg

# Depends on package 'cointReg'.

# Arguments
# x         : a numeric matrix with k * N columns of k I(1) regressors
# y         : a numeric matrix with N columns of N I(1) dependant variables
# deter     : a numeric matrix with p columns of p deterministic regressors
# k         : integer, number of I(1) regressors per cointegrating regression
# N         : integer, number of cointegrating regressions
# kernel    : charcter, any of "ba", "pa", "qs" or "tr", used for kernel based
#             long-run (co-)variance estimation
# bandwidth : character any of "and" or "nw", the bandwidth rule used for
#             kernel estimation
# demeaning : logical, indicates whether or not the series of residuals should
#             be demeaned before computing the long-run (co-)variances
# simpl.var : logical, if TRUE covaraince structure is simplified as specified
#             in Theising and Wied (2023)
# ...       : additional arguments passed to the function getLongRunVar


# Value 
# The function returns a named list with the following entries
# theta.pfm.gls : a (p + k) by 1 numeric matrix containing the PFM-GLS estimate
# theta.fm.sur  : a (p + k) by N numeric matrix containing the FM-SUR estimates
# Omegauv       : the N by (p + k) * N matrix of estimated long-run covariances
#                 u (errors of regression) and v (errors of regressors)
# Omegau.v      : the N by N estimated long-run variance matrix of the modified 
#                 system error
# Omegau.v.inv  : the inverse of Omegau.v
# Omegavv       : the (p + k) * N by (p + k) * N matrix of estimated long-run
#                 variances of v (errors of regressors)
# Omegavv.inv   : the inverse of Omegavv
# yplus.gls     : the GLS modified dependent variables
# simpl.var     : a logical indicating whether the long-run covariance structure
#                 was simplified (TRUE) or no constraints were used (FALSE)

# Author
# Etienne Theising

cointRegSysGLS = function(x, y, deter, k, N, kernel = c("ba", "pa", "qs", "tr"), 
                          bandwidth = c("and", "nw"), demeaning = FALSE, 
                          simpl.var = FALSE, ...) {
  require("cointReg")
  p = ncol(deter)
  Te = nrow(x)
  deter = matrix(deter, ncol = p, nrow = Te)
  
  u.ols = matrix(0, nrow = Te, ncol = N)
  for (n in 1:N) {
    Z = cbind(deter, x[, 1:k + (n - 1) * k])
    u.ols[, n] = lm(y[, n] ~ Z - 1)$residuals
  }
  x.delta = matrixStats::colDiffs(x)
  if (simpl.var) {
    # compute all pairwise correlations and calculate the mean(s)
    Omega.star = matrix(0, nrow = 2 + 2 * k, ncol = 2 + 2 * k)
    Delta.star = matrix(0, nrow = 2 + 2 * k, ncol = 2 + 2 * k)
    for (n in 1:(N - 1)) {
      for (n2 in (n + 1):N) {
        u4var = u.ols[-1, c(n, n2)]
        u = cbind(u4var, x.delta[, c(1:k + (n - 1) * k, 1:k + (n2 - 1) * k)])
        if (!is.numeric(bandwidth)) {
          bw = cointReg::getBandwidth(u, bandwidth = bandwidth, kernel = kernel, 
                                      check = FALSE)
          band = switch(bandwidth, and = "Andrews", nw = "Newey-West")
        }
        else {
          bw = bandwidth
          band = "set by user"
        }
        lrvar <- cointReg::getLongRunVar(u, kernel = kernel, bandwidth = bw, 
                                         demeaning = demeaning, check = FALSE)
        Omega.star = Omega.star + lrvar[[1]] / ((N - 1) * N / 2)
        Delta.star = Delta.star + lrvar[[2]] / ((N - 1) * N / 2)
      }
    }
    # Calculate all matrix blocks
    blockuu = function(A) {
      out = matrix(A[1, 2], nrow = N, ncol = N)
      diag(out) = rep((A[1, 1] + A[2, 2]) / 2, N)
      return(out)
    }
    Omegauu = blockuu(Omega.star)
    Deltauu = blockuu(Delta.star)
    blockvv = function(A) {
      out = matrix(0, nrow = k * N, ncol = k * N)
      out = out + kronecker(upper.tri(matrix(0, nrow = N, ncol = N)), 
                                    A[3:(2 + k), (3 + k):(2 + 2 * k)])
      out = out + kronecker(lower.tri(matrix(0, nrow = N, ncol = N)), 
                                    A[(3 + k):(2 + 2 * k), 3:(2 + k)])
      out = out + kronecker(diag(1, N), 
                            (A[3:(2 + k), 3:(2 + k)] +
                              A[(3 + k):(2 + 2 * k), (3 + k):(2 + 2 * k)]) / 2)
      return(out)
    }
    Omegavv = blockvv(Omega.star)
    Deltavv = blockvv(Delta.star)
    blockuv = function(A) {
      out = matrix(0, nrow = N, ncol = k * N)
      out = out + kronecker(diag(-1, N) + 1,
                            (A[1, (3 + k):(2 + 2 * k), drop = FALSE] + 
                              A[2, 3:(2 + k), drop = FALSE]) / 2)
      out = out + kronecker(diag(N), 
                            (A[1, 3:(2 + k), drop = FALSE] + 
                              A[2, (3 + k):(2 + 2 * k), drop = FALSE]) / 2)
      return(out)
    }
    Omegauv = blockuv(Omega.star)
    Deltauv = blockuv(Delta.star)
    # Omegavu is t(Omegauv), Deltavu is t(Deltauv), so redundant
  } else {
    u4var = u.ols[-1, ]
    u = cbind(u4var, x.delta)
    if (!is.numeric(bandwidth)) {
      bw = cointReg::getBandwidth(u, bandwidth = bandwidth, kernel = kernel, 
                                  check = FALSE)
      band = switch(bandwidth, and = "Andrews", nw = "Newey-West")
    }
    else {
      bw = bandwidth
      band = "set by user"
    }
    lrvar <- cointReg::getLongRunVar(u, kernel = kernel, bandwidth = bw, 
                                     demeaning = demeaning, check = FALSE)
    Omega.star = lrvar[[1]]
    Omegauu = Omega.star[1:N, 1:N]
    Omegavv = Omega.star[1:(k * N) + N, 1:(k * N) + N]
    Omegauv = Omega.star[1:N, 1:(k * N) + N]
    Delta.star = lrvar[[2]]
    Deltauu = Delta.star[1:N, 1:N]
    Deltavv = Delta.star[1:(k * N) + N, 1:(k * N) + N]
    Deltauv = Delta.star[1:N, 1:(k * N) + N]
  }  
  Omegavv.inv = cointReg:::trySolve(Omegavv)
  Omegau.v = Omegauu - Omegauv %*% Omegavv.inv %*% t(Omegauv)
  Omegau.v.inv = cointReg:::trySolve(Omegau.v)
    
  deltahat = numeric(p + k)
  phihat = numeric((p + k) * N)
  delta.right.inv = Omegau.v.inv %*% Omegauv %*% Omegavv.inv
  for (n in 1:N) {
    deltahat[1:k + p] = deltahat[p + 1:k] + 
      t(Deltauv)[1:k + (n - 1) * k, ] %*% t(Omegau.v.inv[n, , drop = FALSE]) - 
      Deltavv[1:k + (n - 1) * k, ] %*% t(delta.right.inv[n, , drop = FALSE])
    phihat[1:k + n * (p + k) - k] = t(Deltauv)[1:k + (n - 1) * k, ] %*%
      t(Omegau.v.inv[n, , drop = FALSE]) - 
      Deltavv[1:k + (n - 1) * k, ] %*% t(delta.right.inv[n, , drop = FALSE])
  }
  
  yplus.gls = matrix(0, nrow = Te, ncol = N)
  yplus.gls[-1, ] = y[-1, ] - t(Omegauv %*% Omegavv.inv %*% t(x.delta))
  denom.gls = matrix(0, nrow = p + k, ncol = p + k)
  numerat.gls = matrix(0, nrow = p + k, ncol = 1)
  Zt = matrix(0, nrow = p + k, ncol = N)
  denom.sur = matrix(0, nrow = (p + k) * N, ncol = (p + k) * N)
  numerat.sur = matrix(0, nrow = (p + k) * N, ncol = 1)
  Zt.sur = matrix(0, nrow = N * (p + k), ncol = N)
  for (te in 2:Te) {
    Zt = rbind(matrix(deter[te, ], nrow = p, ncol = N),
               matrix(x[te, ], nrow = k, ncol = N))
    Zt.sur = matrix(apply(Zt, 2, function(x) 
                                  c(x, rep(0, (p + k) * N)))[1:((p + k) * N^2)],
                    ncol = N)
    
    denom.gls = denom.gls + Zt %*% Omegau.v.inv %*% t(Zt)
    numerat.gls = numerat.gls + Zt %*% Omegau.v.inv %*% yplus.gls[te, ]
    denom.sur = denom.sur + Zt.sur %*% Omegau.v.inv %*% t(Zt.sur)
    numerat.sur = numerat.sur + Zt.sur %*% Omegau.v.inv %*% yplus.gls[te, ]
  }
  denom.gls = cointReg:::trySolve(denom.gls)
  numerat.gls = numerat.gls - Te * deltahat
  theta.gls = denom.gls %*% numerat.gls
  denom.sur = cointReg:::trySolve(denom.sur)
  numerat.sur = numerat.sur - Te * phihat
  theta.sur = denom.sur %*% numerat.sur
  
  out = list(theta.pfm.gls = theta.gls, theta.fm.sur = theta.sur, 
             Omegauv = Omegauv, Omegau.v = Omegau.v, 
             Omegau.v.inv = Omegau.v.inv, Omegavv = Omegavv, 
             Omegavv.inv = Omegavv.inv, yplus.gls = yplus.gls, 
             simpl.var = simpl.var)
  return(out)
}
