# monitorCointSys is a function that performs the proposed 
# PFM-OLS based monitoring procedures in Theising and Wied (2023) of 
# a structural change in a system of homogeneous and cross-sectionally 
# independent cointegrating regressions.

# The function uses some internal functions and code snippets from the R 
# packages 'cointReg' and 'cointmonitoR'.
# Philipp Aschersleben and Martin Wagner (2016). cointReg: Parameter
# Estimation and Inference in a Cointegrating Regression. R package version
# 0.2.0. https://CRAN.R-project.org/package=cointReg
# Philipp Aschersleben, Martin Wagner and Dominik Wied (2016). cointmonitoR:
# Consistent Monitoring of Stationarity and Cointegrating Relationships. R
# package version 0.1.0. https://CRAN.R-project.org/package=cointmonitoR

# Depends on packages 'cointReg', 'cointmonitoR' and 'matrixStats'.
# Depends on own functions 'qSupLim' and 'cointRegSys'.

# Arguments
# y         : a numeric matrix with N columns of N I(1) dependant variables
# x         : a numeric matrix with k * N columns of k I(1) regressors
# m         : numeric, 0 < m < 1, indicating the fraction of the sample size
#             that is break free and will be used for calibration, i.e. 
#             parameter estimation
# trend     : logical, indicating whether a deterministic time trend in addition
#             to an intercept should be included in the cointegrating regression
#             in addition to an intercept
# alpha     : numeric, 0 < alpha < 1, significance level on which the monitoring
#             procedure should be performed
# critvals  : either of a numeric vector containing three critical values 
#             passed by the user or the path to the folder containing csv
#             files of critical values, defaults to NULL assuming that csv files
#             of critical values are in the current working directory
# bandwidth : character any of "and" or "nw", the bandwidth rule used for
#             kernel estimation
# kernel    : charcter, any of "ba", "pa", "qs" or "tr", used for kernel based
#             long-run (co-)variance estimation
# ...       : additional arguments passed to the function getLongRunVar or
#             cointRegSys

# Value
# A named list containing the following objects
# decision   : The test decision of the monitoring procedures
# breakpoint : The detected breakpoint as index of the data vectors
# theta.pfm  : The PFM-OLS estimate of the cointegrating regression based on the
#              calibration period
# res.pfm    : The PFM-OLS residuals of regressing y on x and deterministics
# critvals   : The critical values used during monitoring for H1, H2 and H3
#              based monitoring procedure
# alpha      : The significance level of the monitoring procedure
# h1.pfm     : The values of the detector H1_PFM-OLS
# h2.pfm     : The values of the detector H2_PFM-OLS
# h3.pfm     : The values of the detector H3_PFM-OLS

# Author
# Etienne Theising

monitorCointSys = function(y, x, m, trend, alpha = 0.05, critvals = NULL,
                           bandwidth = c("and", "nw"), 
                           kernel = c("ba", "pa", "qs", "tr"), ...) {
  require("cointReg")
  require("cointmonitoR")
  N = ncol(y)
  k = ncol(x) / N
  Te = nrow(y)
  if (is.numeric(critvals)) {
    if (length(critvals) != 3L) stop('Need to supply three critical values')
  } else {
    critvals = qSupLim(p = 1 - alpha, trend = trend, k = k, m = m, N = N, 
                       path = critvals)
    critvals = as.numeric(critvals[, 6:8])
  }
  deter = cbind(level = rep(1, Te))
  p = 1 + trend
  if (trend) deter = cbind(deter, trend = 1:Te)
  m.idx = floor(m * Te)
  y.m = matrix(y[1:m.idx, ], nrow = m.idx, ncol = N)
  x.m = matrix(x[1:m.idx, ], nrow = m.idx, ncol = k * N)
  deter.m = matrix(deter[1:m.idx, ], nrow = m.idx)
  models = cointRegSys(x = x.m, y = y.m, deter = deter.m, k = k, N = N, 
                       kernel = kernel, bandwidth = bandwidth, ...)
  theta.pfm = models$theta.pfm
  Omegavv.inv.vu.pfm = models$Omegavv.inv.vu
  omega.u.v2 = models$omega.u.v2

  u.plus = matrix(0, nrow = Te, ncol = N)
  for (n in 1:N) {
    x.delta = matrixStats::colDiffs(x[, ((n - 1) * k + 1):(n * k)])
    Z = cbind(deter[-1, ], x[-1, ((n - 1) * k + 1):(n * k)])
    u.plus[, n] = c(0, y[-1, n] - 
                          x.delta %*% Omegavv.inv.vu.pfm - Z %*% theta.pfm)
  }
  h1.pfm = rowSums(
    apply(u.plus[(m.idx + 1):Te, , drop = FALSE], 2, cumsum)^2
    ) / sum(colSums(u.plus[1:m.idx, , drop = FALSE])^2)
  # In h2.pfm's denominator sum(colSums( )) is simplified to just sum
  h2.pfm = rowSums(
    apply((apply(u.plus, 2, cumsum)^2)[(m.idx + 1):Te, , drop = FALSE], 
          2, cumsum)
    ) / sum((apply(u.plus, 2, cumsum)^2)[1:m.idx, , drop = FALSE])
  h3.pfm = rowSums(
    apply(apply(u.plus, 2, cumsum)[(m.idx + 1):Te, , drop = FALSE], 2, cumsum)^2
    ) / sum((colSums(apply(u.plus, 2, cumsum)[1:m.idx, , drop = FALSE]))^2)
  
  svals = (m.idx + 1):Te / Te
  bpoint = numeric(3)
  bpoint[1] = min(c(Te + 1, which(h1.pfm / svals^(2 * p) > critvals[1])))
  bpoint[2] = min(c(Te + 1, which(h2.pfm / svals^(2 * p + 1) > critvals[2])))
  bpoint[3] = min(c(Te + 1, which(h3.pfm / svals^(2 * p + 2) > critvals[3])))
  bpoint = bpoint + m.idx - 1
  bpoint[bpoint > Te] = NA
  out = list(decision = 1 - is.na(bpoint), breakpoint = bpoint,
             theta.pfm = theta.pfm, res.pfm = u.plus,
             critvals = critvals, alpha = alpha,
             h1.pfm = h1.pfm, h2.pfm = h2.pfm, h3.pfm = h3.pfm)
  return(out)
}
