# monitorCointSysGLS is a function that performs the proposed 
# PFM-GLS and FM-SUR based monitoring procedures in Theising and Wied (2023) of 
# a structural change in a system of homogeneous or heterogeneous cointegrating
# regressions with arbitrary error dependence structure.
# There are 3 PFM-GLS based detectors and 2 FM-SUR based detectors used.

# The function uses some internal functions and code snippets from the R 
# packages 'cointReg' and 'cointmonitoR'.
# Philipp Aschersleben and Martin Wagner (2016). cointReg: Parameter
# Estimation and Inference in a Cointegrating Regression. R package version
# 0.2.0. https://CRAN.R-project.org/package=cointReg
# Philipp Aschersleben, Martin Wagner and Dominik Wied (2016). cointmonitoR:
# Consistent Monitoring of Stationarity and Cointegrating Relationships. R
# package version 0.1.0. https://CRAN.R-project.org/package=cointmonitoR

# Depends on packages 'cointReg', 'cointmonitoR', 'matrixStats' and 'parallel'.
# Depends on own functions 'rSupLimGLS', 'qSupLim' and 'cointRegSysGLS'.

# Arguments
# y         : a numeric matrix with N columns of N I(1) dependant variables
# x         : a numeric matrix with k * N columns of k I(1) regressors
# m         : numeric, 0 < m < 1, indicating the fraction of the sample size
#             that is break free and will be used for calibration, i.e. 
#             parameter estimation
# trend     : logical, indicating whether a deterministic time trend in addition
#             to an intercept should be included in the cointegrating regression
#             in addition to an intercept
# alpha     : numeric, 0 < alpha < 1, significance level on which the monitoring
#             procedure should be performed
# critvals  : optional, a numeric vector containing five critical values 
#             passed by the user, defaults to NULL in which case the critical
#             values are obtained via simulation
# bm_prec   : an integer, the number of iid. standard normal random variables 
#             used for a random walk that approximates a Brownian motion
# sim_reps  : an integer, the number of pseudo random numbers to be drawn to 
#             approximate the distribution
# paral     : logical, if TRUE simulation of critical values is parallelized 
#             using a FORK cluster and parSapply with 'parallel' package
# nnodes    : an integer, number of nodes to be forked if computation is 
#             parallel, i.e. paral = TRUE
# bandwidth : character any of "and" or "nw", the bandwidth rule used for
#             kernel estimation
# kernel    : charcter, any of "ba", "pa", "qs" or "tr", used for kernel based
#             long-run (co-)variance estimation
# simpl.var : logical, if TRUE covaraince structure is simplified as specified
#             in Theising and Wied (2023)
# ...       : additional arguments passed to the function getLongRunVar or
#             cointRegSysGLS

# Value
# A named list containing the following objects
# decision      : The test decision of the monitoring procedures
# breakpoint    : The detected breakpoint as index of the data vectors
# theta.pfm.gls : The PFM-GLS estimate of the cointegrating regression based on 
#                 the calibration period
# theta.fm.sur  : The FM-SUR estimates of the cointegrating regressions based on 
#                 the calibration period
# res.pfm.gls   : The PFM-GLS residuals of regressing y on x and deterministics 
# res.fm.sur    : The FM-SUR residuals of regressing y on x and deterministics 
# critvals      : The critical values used during monitoring for H1, H2 and H3
#                 based monitoring procedure
# alpha         : The significance level of the monitoring procedure
# h1.pfm.gls    : The values of the detector H1_PFM-GLS
# h2.pfm.gls    : The values of the detector H2_PFM-GLS
# h3.pfm.gls    : The values of the detector H3_PFM-GLS
# h2.fm.sur     : The values of the detector H2_FM-SUR
# h3.fm.sur     : The values of the detector H3_FM-SUR
# simpl.var     : Logical indicating whether the long-run covariance structure
#                 was simplified (TRUE) or without constraints (FALSE)

# Author
# Etienne Theising

monitorCointSysGLS = function(y, x, m, trend, alpha = 0.05, critvals = NULL,
                              bm_prec = 1000L, sim_reps = 1000L, 
                              paral = FALSE, nnodes = NULL, 
                              bandwidth = c("and", "nw"),
                              kernel = c("ba", "pa", "qs", "tr"),
                              simpl.var = FALSE, ...) {
  require("cointReg")
  require("cointmonitoR")
  require("matrixStats")
  N = ncol(y)
  k = ncol(x) / N
  Te = nrow(y)
  deter = cbind(level = rep(1, Te))
  p = 1 + trend
  if (trend) deter = cbind(deter, trend = 1:Te)
  m.idx = floor(m * Te)
  y.m = matrix(y[1:m.idx, ], nrow = m.idx, ncol = N)
  x.m = matrix(x[1:m.idx, ], nrow = m.idx, ncol = k * N)
  deter.m = matrix(deter[1:m.idx, ], nrow = m.idx)
  models = cointRegSysGLS(x = x.m, y = y.m, deter = deter.m, k = k, N = N, 
                          kernel = kernel, bandwidth = bandwidth,
                          simpl.var = simpl.var, ...)
  sqrtMatrix = function(A) {
    e = eigen(A)
    if (any(e$values < 0)) return (NA)
    V = e$vectors
    A.sqrt = V %*% diag(sqrt(e$values)) %*% t(V)
    return (A.sqrt)
  }
  Omegau.v.sqrt = sqrtMatrix(models$Omegau.v)
  Omegau.v.inv.sqrt = sqrtMatrix(models$Omegau.v.inv)
  Omegavv.sqrt = sqrtMatrix(models$Omegavv)
  if (simpl.var & 
      any(is.na(c(Omegau.v.sqrt, Omegau.v.inv.sqrt, Omegavv.sqrt)))) {
    wmsg = paste(
      "Something wrong with the long-run covariance simplification.",
      "Computation of matrix square roots not possible.", 
      "Resort to long-run covariances without restraints.")
    warning(wmsg)
    simpl.var = FALSE
    models = cointRegSysGLS(x = x.m, y = y.m, deter = deter.m, k = k, N = N, 
                            kernel = kernel, bandwidth = bandwidth,
                            simpl.var = simpl.var, ...)
    Omegau.v.sqrt = sqrtMatrix(models$Omegau.v)
    Omegau.v.inv.sqrt = sqrtMatrix(models$Omegau.v.inv)
    Omegavv.sqrt = sqrtMatrix(models$Omegavv)
    if (any(is.na(c(Omegau.v.sqrt, Omegau.v.inv.sqrt, Omegavv.sqrt)))) {
      ermsg = paste("Something wrong with long-run covariance estimation.",
                    "Not able to compute matrix square root.")
      stop(ermsg)
    }
  }
  theta.pfm.gls = models$theta.pfm.gls
  theta.fm.sur = models$theta.fm.sur
  Omegauv = models$Omegauv
  Omegau.v.inv = models$Omegau.v.inv
  Omegavv.inv = models$Omegavv.inv
  yplus = models$yplus
  
  x.delta = matrixStats::colDiffs(x)
  yplus = rbind(yplus, 
                y[(m.idx + 1):Te, ] - 
                  t(Omegauv %*% Omegavv.inv %*% t(x.delta[m.idx:(Te - 1), ])))
  uplus = matrix(0, nrow = Te, ncol = N)
  uplus.sur = uplus
  for (te in 2:Te) {
    Zt = rbind(matrix(deter[te, ], nrow = p, ncol = N),
               matrix(x[te, ], nrow = k, ncol = N))
    Zt.sur = matrix(
      apply(Zt, 2, function(x) c(x, rep(0, (p + k) * N)))[1:((p + k) * N^2)], 
      ncol = N)
    uplus[te, ] = yplus[te, ] - t(Zt) %*% theta.pfm.gls
    uplus.sur[te, ] = yplus[te, ] - t(Zt.sur) %*% theta.fm.sur
  } 
  # Calculate the detectors
  h1.gls = rowSums2(colCumsums(uplus[(m.idx + 1):Te, , drop = FALSE])^2) / 
    sum(colSums2(uplus[1:m.idx, , drop = FALSE])^2)
  h2.gls = rowSums2(
    colCumsums((colCumsums(uplus)^2)[(m.idx + 1):Te, , drop = FALSE])
    ) / sum((colCumsums(uplus)^2)[1:m.idx, , drop = FALSE])
  h3.gls = rowSums2(
    colCumsums(colCumsums(uplus)[(m.idx + 1):Te, , drop = FALSE])^2
    ) / sum(colSums2(colCumsums(uplus)[1:m.idx, , drop = FALSE])^2)
  h2.sur = rowSums2(
    colCumsums(colCumsums(uplus.sur)^2)[(m.idx + 1):Te, , drop = FALSE]
    ) / sum((colCumsums(uplus.sur)^2)[1:m.idx, , drop = FALSE])
  h3.sur = rowSums2(
    colCumsums(colCumsums(uplus.sur)[(m.idx + 1):Te, , drop = FALSE])^2
    ) / sum(colSums2(colCumsums(uplus.sur[1:m.idx, , drop = FALSE]))^2)
  
  if (is.numeric(critvals)) {
    if (length(critvals) != 5L) stop('Need to supply five critical values')
  } else {
    critvals = qSupLimGLS(p = 1 - alpha, trend = trend, k = k, m = m, N = N, 
                          bm_prec = bm_prec, sim_reps = sim_reps, 
                          Omegau.v.sqrt = Omegau.v.sqrt, 
                          Omegau.v.inv = Omegau.v.inv, 
                          Omegau.v.inv.sqrt = Omegau.v.inv.sqrt,
                          Omegavv.sqrt = Omegavv.sqrt, 
                          paral = paral, nnodes = nnodes)
    critvals = as.numeric(critvals[, 2:6])
  }
  svals = (m.idx + 1):Te / Te
  bpoint = numeric(5)
  bpoint[1] = min(c(Te + 1, which(h1.gls / svals^(2 * p) > critvals[1])))
  bpoint[2] = min(c(Te + 1, which(h2.gls / svals^(2 * p + 1) > critvals[2])))
  bpoint[3] = min(c(Te + 1, which(h3.gls / svals^(2 * p + 2) > critvals[3])))
  bpoint[4] = min(c(Te + 1, which(h2.sur / svals^(2 * p + 1) > critvals[4])))
  bpoint[5] = min(c(Te + 1, which(h3.sur / svals^(2 * p + 2) > critvals[5])))
  bpoint = bpoint + m.idx - 1
  bpoint[bpoint > Te] = NA
  decision = 1 - is.na(bpoint)
  decision[is.na(critvals)] = NA
  out = list(decision = decision, breakpoint = bpoint,
             theta.pfm.gls = theta.pfm.gls, theta.fm.sur = theta.fm.sur, 
             res.pfm.gls = uplus, res.fm.sur = uplus.sur, critvals = critvals, 
             alpha = alpha, h1.pfm.gls = h1.gls, h2.pfm.gls = h2.gls, 
             h3.pfm.gls = h3.gls, h2.fm.sur = h2.sur, h3.fm.sur = h3.sur,
             simpl.var = simpl.var)
  return(out)
}