# qSupLim is a function that returns the p quantile of 
#             \sup_{m <= s <=1} \matchcal{H}^{m,+} (s) / g(s)
# the limiting distribution of any of the test statistics based on PFM-OLS
# estimation.
# The p quantiles are the critical values for the respective significance level
# 1 - p monitoring procedures.

# Arguments
# p       : numeric, probability in {0.900, 0.901, ..., 0.998, 0.999}, 
#           for 0.9 < p < 0.999 outside the above set quantiles are linearly 
#           interpolated
# trend   : logical, indicating whether the cointegrating regression contains 
#           only an intercept or a linear time trend as well
# k       : an integer, the number of I(1) regressors
# m       : a numeric, 0 < m < 1, the length of the calibration period as a 
#           fraction of the sample size
# N       : an integer, the number of cointegrating relationships
# path    : a character, containing the path to csv files of critical values,
#           defaults to NULL assuming critical values are stored in the current
#           working directory

# Value
# A data frame containing the arguments passed to the function and the 
# p quantiles of the above limiting distributions based on PFM-OLS estimation
# and detectors H1, H2 or H3, i.e. the critical values of the respective
# monitoring procedure with significance level 1 - p.

# Author
# Etienne Theising


qSupLim = function(p, trend, k, m, N, path = NULL) {
  if (trend) Dt = "linear" else Dt = "intercept"
  filename = file.path(paste("cv_N", N, "_", Dt, "_PFM-OLS.csv", sep = ""))
  if (is.null(path)) path = "."
  if (!(filename %in% list.files(path))) 
    stop("working directory does not contain critical values")
  if (is.character(path)) filename = file.path(path, filename)
  cv = read.csv(filename)
  p.check = trunc(1000 * p) / 1000
  if (p.check < p) {
    p2 = rep(p.check, 2) + 0:1 / 1000
    lambda = 1000 * (p - p.check)
    out = cv[cv$k == k & cv$m == m & cv$p == p2, ]
    out = cbind(out[1, 1:2], p = p,
                c(1 - lambda, lambda) %*% as.matrix(out[, 4:6]))
  } else out = cv[cv$k == k & cv$m == m & cv$p == p, ]
  out = cbind(trend = trend, N = N, out)
  return (out)
}
