# qSupLimGLS is a function that returns the p quantile of 
#             \sup_{m <= s <=1} \matchcal{H}^{m,+} (s) / g(s)
# the limiting distribution of any of the test statistics based on PFM-GLS
# and FM-SUR estimation.
# The p quantiles are the critical values for the respective significance level
# 1 - p monitoring procedures.
# The distribution approximation of \matchcal{H}^{m,+} (s) depends on the 
# supplied estimates of long-run covariances.

# Depends on pacakge 'parallel' if parallelization on FORK cluster is used.
# Depends on function 'rSupLimGLS'.

# Arguments
# p                 : numeric, vector of probabilities
# trend             : logical, indicating whether the cointegrating regression 
#                     contains only an intercept or a linear time trend as well
# k                 : an integer, the number of I(1) regressors
# m                 : a numeric, 0 < m < 1, the length of the calibration period
#                     as a fraction of the sample size
# N                 : an integer, the number of cointegrating relationships
# bm_prec           : an integer, the number of iid. standard normal random 
#                     variables used for a random walk that approximates a 
#                     Brownian motion
# sim_reps          : an integer, the number of pseudo random numbers to be
#                     drawn to approximate the distribution
# Omegau.v.sqrt     : an N by N numeric matrix, the square root of the long-run 
#                     covariance matrix of the GLS fully modified system error
# Omegau.v.inv      : an N by N numeric matrix, the inverse of the long-run 
#                     covariance matrix of the GLS fully modified system error
# Omegau.v.inv.sqrt : an N by N numeric matrix, the square root of the inverse 
#                     of the long-run covariance matrix of the GLS fully 
#                     modified system error
# Omegavv.sqrt      : an N * k by N * k numeric matrix, the square root of the
#                     long-run covariance of the errors v_t generating the I(1)-
#                     regressors
# paral             : logical, if TRUE simulation is parallelized using a FORK
#                     cluster and parSapply with 'parallel' package
# nnodes            : an integer, number of nodes to be forked if computation
#                     is parallel, i.e. paral = TRUE

# Value
# A data frame containing the p quantiles of the above limiting distributions 
# based on PFM-GLS and FM-SUR estimation and detectors H1, H2 or H3, i.e. the 
# critical values of the respective monitoring procedure with significance level
# 1 - p.

# Author
# Etienne Theising

qSupLimGLS = function(p, trend, k, m, N, bm_prec = 1000L, sim_reps = 1000L,
                      Omegau.v.sqrt, Omegau.v.inv, Omegau.v.inv.sqrt,
                      Omegavv.sqrt, paral = FALSE, nnodes = NULL) {
  iterSim = function() { # wrapper function for iterative simulation
    distr = matrix(NA, nrow = 5, ncol = sim_reps)
    for (i in 1:sim_reps) {
      distr[, i] = rSupLimGLS(trend = trend, k = k, m = m, N = N, 
                              bm_prec = bm_prec, Omegau.v.sqrt = Omegau.v.sqrt,
                              Omegau.v.inv = Omegau.v.inv,
                              Omegau.v.inv.sqrt = Omegau.v.inv.sqrt,
                              Omegavv.sqrt = Omegavv.sqrt)
    }
    return(distr)
  }
  paraSim = function(i) { # wrapper function for parallel simulation
    return(rSupLimGLS(trend = trend, k = k, m = m, N = N, bm_prec = bm_prec, 
                      Omegau.v.sqrt = Omegau.v.sqrt, 
                      Omegau.v.inv = Omegau.v.inv, 
                      Omegau.v.inv.sqrt = Omegau.v.inv.sqrt,
                      Omegavv.sqrt = Omegavv.sqrt))
  }
  if(paral & !is.null(nnodes)){
    require("parallel")
    distr = tryCatch({
      cl = makeCluster(nnodes, type = "FORK")
      distr = parSapply(cl, 1:sim_reps, paraSim)
      stopCluster(cl)
      distr
    },
    warning = function(war) {
      print(paste("Warning:", war))
      iterSim()
    },
    error = function(err) {
      print(paste("Error:", err))
      iterSim()
    }, finally = {})
  } else distr = iterSim()
  # distr is a 5 x sim_reps matrix
  na.returns = is.na(distr)
  if (any(na.returns)) {
    wmsg = paste(
      "At least 1 generated random number discarded due to singularity of", 
      "matrices. Perhaps incorrect covariances supplied.")
    warning(wmsg)
  }
  if (any(rowSums(na.returns) > sim_reps / 100)) {
    wmsg = paste(
      "Something wrong with the simulation of critical values.",
      "More than 1% of generated random numbers discarded due to singularity", 
      "of matrices. Perhaps incorrect covariances supplied. NA returned.")
    warning(wmsg)
    out = data.frame(p, matrix(NA, nrow = length(p), ncol = 5))
    return()
  } else {
    probs = ceiling(p * ncol(distr))
    distr = apply(distr, 1, sort) # now, distr is a sim_reps by 5 matrix
    out = data.frame(p, matrix(distr[probs, ], ncol = 5))
  }
  names(out) = c("p", "H1.PFM.GLS", "H1.PFM.GLS", "H1.PFM.GLS",
                 "H2.FM.SUR", "H3.FM.SUR")
  return (out)
}
