# rSupLimGLS is a function that draws one pseudo random of 
#     \sup_{m <= s <=1} \matchcal{H}^{m,+} (s) / g(s)
# by approximating the limiting distribution \matchcal{H}^{m,+} (s) using 
# functions of random walks as approximation for functionals of Brownian 
# motions. \matchcal{H}^{m,+} (s) is here based on PFM-GLS and FM-SUR
# estimation techniques.
# An estimate of certain long-run variance matrices need to be supplied.
# The function deals with intercept or linear time trend only as deterministic
# regressors but can easily be extended.

# Depends on package 'matrixStats'.

# Arguments
# trend             : logical, indicating whether the cointegrating regression 
#                     contains only an intercept or a linear time trend as well
# k                 : an integer, the number of I(1) regressors
# m                 : a numeric, 0 < m < 1, the length of the calibration period
#                     as a fraction of the sample size
# N                 : an integer, the number of cointegrating relationships
# bm_prec           : an integer, the number of iid. standard normal random 
#                     variables used for a random walk that approximates a 
#                     Brownian motion
# Omegau.v.sqrt     : an N by N numeric matrix, the square root of the long-run 
#                     covariance matrix of the GLS fully modified system error
# Omegau.v.inv      : an N by N numeric matrix, the inverse of the long-run 
#                     covariance matrix of the GLS fully modified system error
# Omegau.v.inv.sqrt : an N by N numeric matrix, the square root of the inverse 
#                     of the long-run covariance matrix of the GLS fully 
#                     modified system error
# Omegavv.sqrt      : an N * k by N * k numeric matrix, the square root of the
#                     long-run covariance of the errors v_t generating the I(1)-
#                     regressors

# Value
# A named numeric vector containing one pseudo random number generated for
# each of the three PFM-GLS based monitoring test statistics and for each
# of the three FM-SUR based monitoring test statistics.

# Author
# Etienne Theising

rSupLimGLS = function(trend, k, m, N, bm_prec = 1000L, 
                      Omegau.v.sqrt, Omegau.v.inv, Omegau.v.inv.sqrt, 
                      Omegavv.sqrt) {
  require("matrixStats")
  mm = floor(m * bm_prec)
  p = 1
  Dt = matrix(1, nrow = 1, ncol = bm_prec)
  Dt_int = seq(1 / bm_prec, 1, by = 1 / bm_prec)
  if (trend) {
    p = 2
    Dt = rbind(Dt, seq(1 / bm_prec, 1, by = 1 / bm_prec))
    Dt_int = rbind(Dt_int, 0.5 * Dt_int^2)
  }
  
  
  W_uv = t(colCumsums(replicate(N, rnorm(bm_prec, sd = sqrt(1 / bm_prec)))))
  W_v = t(colCumsums(replicate(k * N, rnorm(bm_prec, sd = sqrt(1 / bm_prec)))))
  B_v = Omegavv.sqrt %*% W_v
  
  JOJ_int = matrix(0, nrow = p + k, ncol = p + k)
  JOdW_int = matrix(0, nrow = p + k, ncol = 1)
  JOJ_int_sur = matrix(0, nrow = (p + k) * N, ncol = (p + k) * N)
  JOdW_int_sur = matrix(0, nrow = (p + k) * N, ncol = 1)
  for (te in 1:(mm - 1)) {
    J = rbind(matrix(Dt[, te], nrow = p, ncol = N), 
              matrix(B_v[, te], nrow = k, ncol = N))
    JOJ_int = JOJ_int + J %*% Omegau.v.inv %*% t(J)
    JOdW_int = JOdW_int + J %*% Omegau.v.inv.sqrt %*% 
                                                  (W_uv[, te + 1] - W_uv[, te])
    Jbold = matrix(apply(J, 2, function(x) 
                                  c(x, rep(0, (p + k) * N)))[1:((p + k) * N^2)],
                   ncol = N)
    JOJ_int_sur = JOJ_int_sur + Jbold %*% Omegau.v.inv %*% t(Jbold)
    JOdW_int_sur = JOdW_int_sur + Jbold %*% Omegau.v.inv.sqrt %*% 
                                                  (W_uv[, te + 1] - W_uv[, te])
  }
  J = rbind(matrix(Dt[, mm], nrow = p, ncol = N), 
            matrix(B_v[, mm], nrow = k, ncol = N))
  JOJ_int = JOJ_int + J %*% Omegau.v.inv %*% t(J)
    
  Jbold = matrix(apply(J, 2, function(x) 
                                  c(x, rep(0, (p + k) * N)))[1:((p + k) * N^2)],
                 ncol = N)
  JOJ_int_sur = JOJ_int_sur + Jbold %*% Omegau.v.inv %*% t(Jbold)
    
  int.m = tryCatch(expr = {bm_prec * solve(JOJ_int) %*% JOdW_int},
                   warning = function(war) {
                     print(paste("Warning:", war))
                     FALSE
                   },
                   error = function(err) {
                     print(paste("Error:", err))
                     FALSE
                   },
                   finally = {})
  if (!is.logical(int.m)) {
    int.m.sur = tryCatch(expr = {bm_prec * solve(JOJ_int_sur) %*% JOdW_int_sur},
                         warning = function(war) {
                           print(paste("Warning:", war))
                           FALSE
                         },
                         error = function(err) {
                           print(paste("Error:", err))
                           FALSE
                         },
                         finally = {})
  error = is.logical(int.m.sur)
  } else {
    error = TRUE
  }
  
  if (error) {
    wmsg = paste(
      "Something wrong with the simulation of critical values.",
      "More than 1% of generated random numbers discarded due to singularity",
      "of matrices.",
      "Perhaps incorrect covariances supplied.")
    warning(wmsg)
    out = rep(NA, 6)
    names(out) = c("H1.PFM.GLS", "H2.PFM.GLS", "H3.PFM.GLS",
                   "H1.FM.SUR", "H2FM.SUR", "H3.FM.SUR")
    return(NA)
  }
  
  B_v.int = rowCumsums(B_v) / bm_prec
  int.s = t(matrix(
    apply(rbind(Dt_int, B_v.int), 2, function(x)
      rbind(matrix(x[1:p], nrow = p, ncol = N), matrix(x[1:(N * k) + p], ncol = N))
    ), nrow = (p + k))) # int.s is now N*bm_prec x p+k matrix
  int.s.sur = t(matrix(
    apply(rbind(Dt_int, B_v.int), 2, function(x)
      apply(rbind(matrix(x[1:p], nrow = p, ncol = N), 
                  matrix(x[1:(N * k) + p], ncol = N)), 
            MARGIN = 2,
            function(y) c(y, rep(0, (p + k) * N))
      )[1:((p + k) * N^2)]
    ), nrow = (p + k) * N)) # int.s.sur is now N*bm_prec x N*(p+k) matrix
  
  Omegau.v.sqrt_W_uv = Omegau.v.sqrt %*% W_uv
  minuend = matrix(int.s %*% int.m, nrow = N, ncol = bm_prec)
  minuend.sur = matrix(int.s.sur %*% int.m.sur, nrow = N, ncol = bm_prec)
  W.pfm.gls = Omegau.v.sqrt_W_uv - minuend
  W.fm.sur = Omegau.v.sqrt_W_uv - minuend.sur
  
  W.pfm.gls0m = matrix(W.pfm.gls[, 1:mm], nrow = N)
  W.fm.sur0m = matrix(W.fm.sur[, 1:mm], nrow = N)
  
  W.pfm.gls = matrix(W.pfm.gls[, (mm + 1):bm_prec], nrow = N)
  W.fm.sur = matrix(W.fm.sur[, (mm + 1):bm_prec], nrow = N)
  
  # now, approximate the limiting distributions of the different 
  # PFM-GLS and FM-SUR based detectors
  H1.pfm = colSums((W.pfm.gls - W.pfm.gls0m[, mm])^2) / sum(W.pfm.gls0m[, mm]^2)
  
  # note that in the demoninator the integral approximation and sum_n_to_N 
  # is done via one application of the sum command
  H2.pfm = colSums(rowCumsums(W.pfm.gls^2)) / sum(W.pfm.gls0m^2)
  H2.sur = colSums(rowCumsums(W.fm.sur^2)) / sum(W.fm.sur0m^2)
  
  H3.pfm = colSums(rowCumsums(W.pfm.gls)^2) / sum(rowSums(W.pfm.gls0m)^2)
  H3.sur = colSums(rowCumsums(W.fm.sur)^2) / sum(rowSums(W.fm.sur0m)^2)
  
  grid = seq((mm + 1) / bm_prec, 1, by = 1 / bm_prec)
  return (c(H1.PFM.GLS = max(H1.pfm / grid^(2 * p)), 
            H2.PFM.GLS = max(H2.pfm / grid^(2 * p + 1)), 
            H3.PFM.GLS = max(H3.pfm / grid^(2 * p + 2)), 
            H2.FM.SUR = max(H2.sur / grid^(2 * p + 1)), 
            H3.FM.SUR = max(H3.sur / grid^(2 * p + 2))
  ))
}